;; Generated package description from vc-got.el  -*- no-byte-compile: t -*-
(define-package "vc-got" "1.1.1" "VC backend for Game of Trees VCS" '((emacs "25.1")) :commit "18cc42ab3c12d124fb013923f4583dd25ffe5dcd" :authors '(("Omar Polo" . "op@omarpolo.com") ("Timo Myyrä" . "timo.myyra@bittivirhe.fi")) :maintainer '("Omar Polo" . "op@omarpolo.com") :keywords '("vc" "tools") :url "https://projects.omarpolo.com/vc-got.html")
